package ru.dbotthepony.mc.otm.capability.drive;

import net.minecraft.MethodsReturnNonnullByDefault;
import net.minecraft.nbt.CompoundTag;
import net.minecraft.nbt.ListTag;
import net.minecraft.nbt.Tag;
import ru.dbotthepony.mc.otm.core.Fraction;
import ru.dbotthepony.mc.otm.storage.*;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.annotation.ParametersAreNonnullByDefault;
import java.math.BigDecimal;
import java.util.*;

@MethodsReturnNonnullByDefault
@ParametersAreNonnullByDefault
abstract public class AbstractMatteryDrive<T extends IStorageStack> implements IMatteryDrive<T> {
	protected final HashMap<Object, List<IStorageTuple<T>>> items = new HashMap<>();
	protected final HashMap<UUID, IStorageTuple<T>> items_by_id = new HashMap<>();

	abstract public StorageObjectTuple<T> identity();

	protected boolean dirty = false;
	protected int different_stacks = 0;
	protected Fraction stored = Fraction.ZERO;
	protected int max_different_stacks;
	protected Fraction capacity;

	public AbstractMatteryDrive(Fraction capacity, int max_different_stacks) {
		this.capacity = capacity;
		this.max_different_stacks = max_different_stacks;
	}

	public AbstractMatteryDrive(Fraction capacity) {
		this(capacity, 0xFFFF);
	}

	@Override
	public boolean isDirty() {
		return dirty;
	}

	@Override
	public void markDirty() {
		dirty = true;
	}

	@Override
	public void markClean() {
		dirty = false;
	}

	@Override
	public Fraction getStoredCount() {
		return stored;
	}

	@Override
	public Fraction getCapacity() {
		return capacity;
	}

	@Nonnull
	@Override
	@SuppressWarnings("unchecked")
	public T insertStack(T stack, boolean simulate) {
		var max_insert = getCapacity().minus(getStoredCount()).min(stack.getCount());

		if (max_insert.compareTo(Fraction.ZERO) <= 0)
			return stack;

		final var listing = items.computeIfAbsent(stack.partitionKey(), (key) -> new ArrayList<>());

		for (var state : listing) {
			if (state.stack().sameItem(stack)) {
				if (!simulate) {
					state.stack().grow(max_insert);
					stored = stored.plus(max_insert);

					for (var listener : listeners2) {
						listener.changeObject(state.id(), state.stack().getCount());
					}

					markDirty();
				}

				final var copy_item = (T) stack.copy();
				copy_item.shrink(max_insert);
				return copy_item;
			}
		}

		if (different_stacks >= max_different_stacks) {
			return stack;
		}

		if (!simulate) {
			different_stacks++;
			stored = stored.plus(max_insert);

			final var copy = (T) stack.copy();
			copy.setCount(max_insert);
			final var state = new StorageTuple<>(UUID.randomUUID(), copy);
			listing.add(state);
			items_by_id.put(state.id(), state);

			for (var listener : listeners2) {
				listener.addObject(state.stack(), state.id(), this);
			}

			markDirty();
		}

		final var copy_item = (T) stack.copy();
		copy_item.shrink(max_insert);
		return copy_item;
	}

	@Nonnull
	@Override
	@SuppressWarnings("unchecked")
	public T extractStack(UUID id, Fraction amount, boolean simulate) {
		var get = items_by_id.get(id);

		if (get == null)
			return identity().empty();

		if (amount.compareTo(Fraction.ZERO) <= 0)
			amount = get.stack().getMaxStackSize().orElse(get.stack().getCount());

		amount = amount.min(get.stack().getCount());

		if (amount.compareTo(Fraction.ZERO) <= 0)
			return identity().empty();

		final var copy = (T) get.stack().copy();
		copy.setCount(amount);

		if (!simulate) {
			if (amount.compareTo(get.stack().getCount()) == 0) {
				var listing = items.get(get.stack().partitionKey());
				listing.remove(get);
				different_stacks--;

				for (var listener : listeners2) {
					listener.removeObject(get.id());
				}

				if (listing.size() == 0) {
					items.remove(get.stack().partitionKey());
				}
			}

			stored = stored.minus(amount);
			get.stack().shrink(amount);

			if (get.stack().getCount().compareTo(Fraction.ZERO) != 0) {
				for (var listener : listeners2) {
					listener.changeObject(get.id(), get.stack().getCount());
				}
			}

			markDirty();
		}

		return copy;
	}

	@Nullable
	abstract protected CompoundTag serializeStack(IStorageTuple<T> item);

	@Override
	public CompoundTag serializeNBT() {
		final var compound = new CompoundTag();

		compound.putString("capacity", capacity.toString());
		compound.putInt("max_different_stacks", max_different_stacks);

		var list_items = new ListTag();
		compound.put("items", list_items);

		for (var entry : items.entrySet()) {
			for (var stack : entry.getValue()) {
				CompoundTag entry_compound;

				if ((entry_compound = serializeStack(stack)) != null) {
					entry_compound.putLongArray("id", new long[] { stack.id().getMostSignificantBits(), stack.id().getLeastSignificantBits() });
					list_items.add(entry_compound);
				}
			}
		}

		return compound;
	}

	@Nullable
	abstract protected T deserializeStack(CompoundTag tag);

	@Override
	public void deserializeNBT(CompoundTag nbt) {
		items.clear();
		items_by_id.clear();
		stored = Fraction.ZERO;
		different_stacks = 0;

		if (nbt.contains("capacity"))
			capacity = Fraction.deserializeNBT(nbt.get("capacity"));

		max_different_stacks = nbt.getInt("max_different_stacks");

		for (var _entry : nbt.getList("items", Tag.TAG_COMPOUND)) {
			if (_entry instanceof CompoundTag entry) {
				final var stack = deserializeStack(entry);

				if (stack != null) {
					stored = stored.plus(stack.getCount());
					different_stacks++;
					final var id = entry.getLongArray("id");

					final var tuple = new StorageTuple<>(id.length == 2 ? new UUID(id[0], id[1]) : UUID.randomUUID(), stack);
					items.computeIfAbsent(stack.partitionKey(), (key) -> new ArrayList<>()).add(tuple);
					items_by_id.put(tuple.id(), tuple);
				}
			}
		}
	}

	@Override
	public Class<T> storageIdentity() {
		return identity().identity();
	}

	@Override
	public T getStack(UUID id) {
		var get = items_by_id.get(id);
		return get == null ? identity().empty() : get.stack();
	}

	@Override
	public List<IStorageTuple<T>> getStacks() {
		int amount = 0;

		for (var listing : items.values())
			amount += listing.size();

		var list = new ArrayList<IStorageTuple<T>>(amount);

		for (var listing : items.values()) {
			list.addAll(listing);
		}

		return list;
	}

	protected final HashSet<IStorageListener<T>> listeners2 = new HashSet<>();

	@Override
	public boolean addListener(IStorageListener<T> listener) {
		return listeners2.add(listener);
	}

	@Override
	public boolean removeListener(IStorageListener<T> listener) {
		return listeners2.remove(listener);
	}
}
