package ru.dbotthepony.mc.otm.item;

import net.minecraft.ChatFormatting;
import net.minecraft.core.Direction;
import net.minecraft.nbt.CompoundTag;
import net.minecraft.network.chat.Component;
import net.minecraft.network.chat.TranslatableComponent;
import net.minecraft.world.item.*;
import net.minecraft.world.level.Level;
import net.minecraftforge.common.capabilities.Capability;
import net.minecraftforge.common.capabilities.ICapabilityProvider;
import net.minecraftforge.common.util.LazyOptional;
import net.minecraftforge.energy.CapabilityEnergy;
import ru.dbotthepony.mc.otm.OverdriveThatMatters;
import ru.dbotthepony.mc.otm.capability.IMatteryEnergyStorage;
import ru.dbotthepony.mc.otm.capability.MatteryCapability;
import ru.dbotthepony.mc.otm.core.Fraction;
import ru.dbotthepony.mc.otm.menu.FormattingHelper;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.util.List;

public class ItemBattery extends Item {
	public class BatteryMatteryCapability implements IMatteryEnergyStorage, ICapabilityProvider {
		private final Fraction storage;
		private final Fraction max_input;
		private final Fraction max_output;
		private final boolean is_creative;

		private final ItemStack stack;

		private Fraction energy() {
			CompoundTag tag = stack.getOrCreateTag();

			if (tag.contains("otm_energy")) {
				return Fraction.deserializeNBT(tag.get("otm_energy"));
			}

			return Fraction.ZERO;
		}

		private void energy(Fraction value) {
			stack.getOrCreateTag().put("otm_energy", value.serializeNBT());
		}

		public BatteryMatteryCapability(ItemStack stack, Fraction storage, Fraction max_input, Fraction max_output) {
			this.stack = stack;
			is_creative = false;
			this.storage = storage;
			this.max_input = max_input;
			this.max_output = max_output;
		}

		public BatteryMatteryCapability(ItemStack stack) {
			this.stack = stack;
			is_creative = true;
			storage = MatteryCapability.LONG_MAX_VALUE;
			max_input = MatteryCapability.LONG_MAX_VALUE;
			max_output = MatteryCapability.LONG_MAX_VALUE;
		}

		@Nonnull
		@Override
		public Fraction extractEnergyOuter(@Nonnull Fraction howMuch, boolean simulate) {
			return extractEnergyInner(howMuch, simulate);
		}

		@Nonnull
		@Override
		public Fraction extractEnergyInner(@Nonnull Fraction howMuch, boolean simulate) {
			if (is_creative)
				return howMuch;

			Fraction new_energy = energy().minus(howMuch.min(max_output)).moreThanZero();
			Fraction diff = energy().minus(new_energy);

			if (!simulate) {
				energy(new_energy);
			}

			return diff;
		}

		@Nonnull
		@Override
		public Fraction receiveEnergyOuter(@Nonnull Fraction howMuch, boolean simulate) {
			return receiveEnergyInner(howMuch, simulate);
		}

		@Nonnull
		@Override
		public Fraction receiveEnergyInner(@Nonnull Fraction howMuch, boolean simulate) {
			if (is_creative)
				return howMuch;

			Fraction new_energy = energy().plus(howMuch.min(max_input)).min(storage);
			Fraction diff = new_energy.minus(energy());

			if (!simulate) {
				energy(new_energy);
			}

			return diff;
		}

		@Nonnull
		@Override
		public Fraction getMissingPower() {
			if (is_creative)
				return MatteryCapability.LONG_MAX_VALUE;

			return IMatteryEnergyStorage.super.getMissingPower();
		}

		@Nonnull
		@Override
		public Fraction getBatteryLevel() {
			if (is_creative)
				return MatteryCapability.LONG_MAX_VALUE;

			return energy();
		}

		@Nonnull
		@Override
		public Fraction getMaxBatteryLevel() {
			if (is_creative)
				return MatteryCapability.LONG_MAX_VALUE;

			return storage;
		}

		private final LazyOptional<IMatteryEnergyStorage> resolver = LazyOptional.of(() -> this);

		@Nonnull
		@Override
		public <T> LazyOptional<T> getCapability(@Nonnull Capability<T> cap, @Nullable Direction side) {
			if (cap == MatteryCapability.ENERGY || cap == CapabilityEnergy.ENERGY)
				return resolver.cast();

			return LazyOptional.empty();
		}

		@Override
		public boolean canExtract() {
			return true;
		}

		@Override
		public boolean canReceive() {
			return true;
		}
	}

	private final boolean is_creative;

	public final Fraction storage;
	public final Fraction receive;
	public final Fraction extract;

	private final Component THROUGHPUT;
	private static final Component INFINITE_STORAGE = new TranslatableComponent("otm.item.power.infinite.storage").withStyle(ChatFormatting.GRAY);

	public ItemBattery(Fraction storage, Fraction receive, Fraction extract) {
		super(new Properties().stacksTo(1).tab(OverdriveThatMatters.CREATIVE_TAB));
		is_creative = false;
		this.storage = storage;
		this.receive = receive;
		this.extract = extract;
		THROUGHPUT = new TranslatableComponent("otm.item.power.normal.throughput", FormattingHelper.formatPower(receive), FormattingHelper.formatPower(extract)).withStyle(ChatFormatting.GRAY);
	}

	public ItemBattery() {
		super(new Properties().stacksTo(1).rarity(Rarity.EPIC).tab(OverdriveThatMatters.CREATIVE_TAB));
		is_creative = true;
		this.storage = MatteryCapability.LONG_MAX_VALUE;
		this.receive = MatteryCapability.LONG_MAX_VALUE;
		this.extract = MatteryCapability.LONG_MAX_VALUE;
		THROUGHPUT = new TranslatableComponent("otm.item.power.infinite.throughput").withStyle(ChatFormatting.GRAY);
	}

	@Override
	public void appendHoverText(ItemStack stack, @Nullable Level p_41422_, List<Component> p_41423_, TooltipFlag p_41424_) {
		super.appendHoverText(stack, p_41422_, p_41423_, p_41424_);

		if (is_creative) {
			p_41423_.add(INFINITE_STORAGE);
			p_41423_.add(THROUGHPUT);
		} else {
			stack.getCapability(MatteryCapability.ENERGY).ifPresent(cap ->
				p_41423_.add(new TranslatableComponent("otm.item.power.normal.storage", FormattingHelper.formatPower(cap.getBatteryLevel()), FormattingHelper.formatPower(cap.getMaxBatteryLevel())).withStyle(ChatFormatting.GRAY))
			);

			p_41423_.add(THROUGHPUT);
		}
	}

	@Nullable
	@Override
	public ICapabilityProvider initCapabilities(ItemStack stack, @Nullable CompoundTag nbt) {
		if (is_creative)
			return new BatteryMatteryCapability(stack);

		return new BatteryMatteryCapability(stack, storage, receive, extract);
	}
}
